#! /usr/bin/python

import sys
from scapy.all import *
from optparse import OptionParser
from random import randint

"""
Author: Demetris Papapetrou and QSecure
Creation Date: one day in 2010
Release Date: 25/09/2014
Purpose: Proof of Concept (PoC) tool for identifying if a host/router generates incremental IP IDs in IP, TCP, UDP, ICMP packets
Dependencies: Scapy network packet generator

Comments: I know that my programming skills suck :-)

License
--------
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
"""

# Turn down the verbosity of scapy
conf.verb = 0

 
def set_parse_options():

	parser = OptionParser(usage="usage: %prog [options] target")
	parser.add_option("-T", dest="tcp",
                  help="use TCP scan mode with one or more of the following flags S=Syn, A=Ack, F=Fin, X=Xmas, N=Null" )
	parser.add_option("-U", dest="udp", action="store_true",
                  help="use UDP scan mode" )
	parser.add_option("-I", dest="icmp",
                  help="use ICMP scan mode P=Ping (default) T=Timestamp A=Address Mask" )
	parser.add_option("-P",dest="ip", action="store_true",
                  help="use IP Protocol scan mode" )
	parser.add_option("-t",dest="tcp_port",
                  help="TCP port to use for scanning" )
	parser.add_option("-u",dest="udp_port",
                  help="UDP port to use for scanning" )
	parser.add_option("-p",dest="ip_proto",
                  help="IP protocol number to use for scanning")
	parser.add_option("--ttl", dest="ttl",
                  help="how many hops the packet will traverse")
	parser.add_option("--tracert", dest="tracert", action="store_true",
                  help="detect IPID for every hop in the path")
	parser.add_option("--max_hops", dest="max_hops",
                  help="terminate traceroute at hop N. (use with --tracert only)")
	parser.add_option("--max_noresp", dest="max_noresp",
                  help="terminate traceroute after N irresponsive hops. (use with --tracert only)")
	parser.add_option("-c", dest="packet_count",
                  help="number of packets to send")

	parser.set_defaults(tcp = None,
		udp = False,
		icmp = None,
		ip = False,
		tracert = False,
		tcp_port = 80,
		udp_port = 161,
		ip_proto = 16,
		packet_count = 1,
		ttl = 128,
		max_hops = 10,
		max_noresp = 3)

	return parser



if __name__ == "__main__":

	parser = set_parse_options()
	(option, args) = parser.parse_args()
	
	#### Section for checking the validity of program parameters ####
	if len(args) != 1:
		parser.print_help()
        	parser.error("incorrect number of arguments")
		sys.exit(2)
	
	try:
		if option.packet_count: 
			count = int(option.packet_count)
	except:
		parser.print_help()
		parser.error("Packet count must be an integer")
	
	try:
		if option.tcp: 
			tcp = str(option.tcp)
	except:
		parser.print_help()
		parser.error("TCP scan mode must contain one or more of the flags S/A/F/X/N")
	
	try:
		if option.icmp: 
			icmp = str(option.icmp)
	except:
		parser.print_help()
		parser.error("ICMP scan mode must contain one of the following option P/T/A")
		
	try:
		if option.tcp_port: 
			tcp_port = int(option.tcp_port)
	except:
		parser.print_help()
		parser.error("TCP port must be an integer")
		
	try:
		if option.udp_port: 
			udp_port = int(option.udp_port)
	except:
		parser.print_help()
		parser.error("UDP port must be an integer")
	
	try:
		if option.ip_proto: 
			ip_proto = int(option.ip_proto)
	except:
		parser.print_help()
		parser.error("IP proto must be an integer < 255")
	
	try:
		if option.ttl: 
			ip_ttl = int(option.ttl)
	except:
		parser.print_help()
		parser.error("TTL value must be an integer < 255")
	
	try:
		if option.max_hops: 
			max_hops = int(option.max_hops)
	except:
		parser.print_help()
		parser.error("max_hops must be an integer <= 64")
		
	try:
		if option.max_noresp: 
			max_noresp = int(option.max_noresp)
	except:
		parser.print_help()
		parser.error("max_noresp must be an integer <= 16")
	
	#### PARAMETER CHECK ####
	if count >= 256:
		print "count must be an integer < 256"
		sys.exit(2)
	
	if tcp_port >= 65536 or udp_port >= 65536:
		print "UDP/TCP ports must be an integer < 65536"
		sys.exit(2)
	
	if ip_proto >= 255:
		print "IP proto must be an integer < 255"
		sys.exit(2)
	
	if ip_ttl >= 255:
		print "TTL value must be an integer < 255"
		sys.exit(2)
	
	#Check that max_hops is 0-64 and used with --tracert
	if option.tracert == True and max_hops > 0 and max_hops <= 64:
		pass
	elif max_hops == 10:
		pass
	else:
		parser.print_help()
		parser.error("max_hops must be an integer <= 64 and must be used with --tracert")
		sys.exit(2)
	
	#Check that max_noresp is 0-16 and used with --tracert
	if option.tracert == True and max_noresp > 0 and max_noresp <= 16:
		pass
	elif max_noresp == 3:
		pass
	else:
		parser.print_help()
		parser.error("max_noresp must be an integer <= 16 and must be used with --tracert")
		sys.exit(2)
		
	#Check if only the allowed options can be set for ICMP scanning
	if option.icmp != None and icmp.find("P") ==-1 and icmp.find("T") ==-1 and icmp.find("A") ==-1:
		parser.print_help()
		parser.error("ICMP scan mode must contain one of the following option P/T/A")
		sys.exit(2)	
	
	#Check if only the allowed flags can be set for TCP scanning
	if option.tcp != None and tcp.find("S") ==-1 and tcp.find("A") ==-1 and tcp.find("F") ==-1 and tcp.find("X") ==-1 and tcp.find("N") ==-1:
		parser.print_help()
		parser.error("TCP scan mode must contain one or more of the flags S/A/F/X/N")
		sys.exit(2)
	
	#If no scan mode is defined default to ICMP	
	if option.tcp == None and option.udp == False and option.icmp == None and option.ip == False:
		option.icmp = "P"
		icmp = "P"
	
		
	
	print "Scan process initiated..."
	##########  TCP  ##########
	try:
		if option.tcp != None:
			#TCP Syn scan
			if tcp.find("S") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags="S"),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP SYN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "TCP SYN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "TCP SYN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
									else:
										print "Error: Cannot identify response"
									previousID = r[IP].id
							for s in unans:
								print "TCP SYN " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP SYN " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags="S"),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP SYN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP SYN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP SYN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP SYN: No Response"
				print "----------------------------------------------------------------------"	
			#TCP Syn/Ack scan
			if tcp.find("SA") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags="SA"),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP SYN/ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "TCP SYN/ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "TCP SYN/ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
									previousID = r[IP].id
							for s in unans:
								print "TCP SYN/ACK " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP SYN/ACK " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags="SA"),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP SYN/ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP SYN/ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP SYN/ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP SYN/ACK: No Response"
				print "----------------------------------------------------------------------"	
			#TCP Ack scan
			if tcp.find("A") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ack_scan_hop = 0
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags="A"),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
										previousID = r[IP].id
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										ack_scan_hop = 1
										ttl_ip = r[IP].src
										ttl_id = r[IP].id
										src_addr = r.getlayer(IP).src
										ans,unans = sr(IP(dst=src_addr,id=RandShort())/TCP(sport=RandShort(),dport=tcp_port,flags="A"),timeout=2)
										for s,r in ans:
											if r.haslayer(TCP):
												if s[TCP].dport == r[TCP].sport:
													print "TCP ACK " + str(hopcount) + ":  ip=" + ttl_ip + " 	ttl_id=" + str(ttl_id) + " 	inc=" + str(ttl_id - previousID) + " 	ack_id=" + str(r[IP].id)  + " 	(TimeX)"
										previousID = ttl_id
									elif r.haslayer(ICMP):
										print "TCP ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
										previousID = r[IP].id
							for s in unans:
								if ack_scan_hop == 1:
									print "TCP ACK " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								else:
									print "TCP ACK " + str(hopcount) + ": No Response"
									no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP ACK " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags="A"),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP ACK:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP ACK: No Response"
				print "----------------------------------------------------------------------"	
			#TCP Fin scan
			if tcp.find("F") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags="F"),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP FIN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "TCP FIN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "TCP FIN " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
									previousID = r[IP].id
							for s in unans:
								print "TCP FIN " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP FIN " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags="F"),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP FIN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP FIN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP FIN:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP FIN: No Response"
				print "----------------------------------------------------------------------"	
			#TCP Xmas scan
			if tcp.find("X") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags="FPU"),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP XMAS " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "TCP XMAS " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "TCP XMAS " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
									previousID = r[IP].id
							for s in unans:
								print "TCP XMAS " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP XMAS " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags="FPU"),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP XMAS:  ip=" + r[IP].src + "	 id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP XMAS:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP XMAS:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP XMAS: No Response"
				print "----------------------------------------------------------------------"	
			#TCP Null scan
			if tcp.find("N") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/TCP(sport=RandShort(),dport=tcp_port,flags=None),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(TCP):
										if s[TCP].dport == r[TCP].sport:
											print "TCP NULL " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
											morehops = False
									elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "TCP NULL " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "TCP NULL " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
										no_response_count = no_response_count + 1
									previousID = r[IP].id
							for s in unans:
								print "TCP NULL " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "TCP NULL " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags=None),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(TCP):
									if s[TCP].dport == r[TCP].sport:
										print "TCP NULL:  ip=" + r[IP].src + "	 id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "TCP NULL:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "TCP NULL:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "TCP NULL: No Response"
				print "----------------------------------------------------------------------"	
		
		##########  UDP  ##########
		if option.udp == True:  
			#UDP scan
			previousID = 0
			if option.tracert:
				morehops = True
				hopcount = 1
				no_response_count = 0
				while morehops:
					previousID = 0
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/UDP(sport=RandShort(),dport=udp_port),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
									print "UDP Scan " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
									morehops = False
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "UDP Scan " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "UDP Scan " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "UDP Scan " + str(hopcount) + ": No Response"
							no_response_count = no_response_count + 1
					if no_response_count >= max_noresp * count:
						print "UDP Scan " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
						morehops = False 
						break
					print "----------------------------------------------------------------------"
					if hopcount == max_hops:
						break
					hopcount = hopcount + 1
			else:
				for i in range(count):
					ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/UDP(sport=RandShort(),dport=udp_port),timeout=2)
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
								print "UDP Scan:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
							if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
								print "UDP Scan:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
							previousID = r[IP].id
					for s in unans:
						print "UDP Scan: No Response"
			print "----------------------------------------------------------------------"	
		
		##########  ICMP  ##########
		if option.icmp != None:
			#ICMP ping scan
			if icmp.find("P") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/ICMP(type=8,code=0),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 0:
										print "ICMP Ping " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
										morehops = False
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "ICMP Ping " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "ICMP Ping " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
									previousID = r[IP].id
							for s in unans:
								print "ICMP Ping " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "ICMP Ping " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/ICMP(type=8,code=0),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 0:
									print "ICMP Ping:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "ICMP Ping:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
									print "ICMP Ping:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								elif r.haslayer(ICMP):
									print "ICMP Ping:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "ICMP Ping: No Response"
				print "----------------------------------------------------------------------"		
			#ICMP timestamp scan
			if icmp.find("T") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/ICMP(type=13,code=0),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 14:
										print "ICMP Timestamp " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
										morehops = False
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "ICMP Timestamp " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "ICMP Timestamp " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
									previousID = r[IP].id
							for s in unans:
								print "ICMP Timestamp " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "ICMP Timestamp " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/ICMP(type=13,code=0),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 14:
									print "ICMP Timestamp:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "ICMP Timestamp:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
									print "ICMP Timestamp:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								elif r.haslayer(ICMP):
									print "ICMP Timestamp:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "ICMP Timestamp: No Response"
				print "----------------------------------------------------------------------"		
			#ICMP address mask scan
			if icmp.find("A") >= 0:
				previousID = 0
				if option.tracert:
					morehops = True
					hopcount = 1
					no_response_count = 0
					while morehops:
						previousID = 0
						for i in range(count):
							ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=(hopcount))/ICMP(type=17,code=0),timeout=2)
							for s,r in ans:
								if r.haslayer(IP):
									if previousID == 0: previousID = r[IP].id
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 18:
										print "ICMP Address Mask " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
										morehops = False
									if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
										print "ICMP Address Mask " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
									elif r.haslayer(ICMP):
										print "ICMP Address Mask " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
									previousID = r[IP].id
							for s in unans:
								print "ICMP Address Mask " + str(hopcount) + ": No Response"
								no_response_count = no_response_count + 1
						if no_response_count >= max_noresp * count:
							print "ICMP Address Mask " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
							morehops = False 
							break
						print "----------------------------------------------------------------------"
						if hopcount == max_hops:
							break
						hopcount = hopcount + 1
				else:
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),ttl=ip_ttl)/ICMP(type=17,code=0),timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 18:
									print "ICMP Address Mask:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "ICMP Address Mask:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
									print "ICMP Address Mask:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								elif r.haslayer(ICMP):
									print "ICMP Address Mask:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								previousID = r[IP].id
						for s in unans:
							print "ICMP Address Mask: No Response"
				print "----------------------------------------------------------------------"	
		
		##########  IP  ##########
		if option.ip == True:
			#IP protocol scan
			previousID = 0
			if option.tracert:
				morehops = True
				hopcount = 1
				no_response_count = 0
				while morehops:
					previousID = 0
					for i in range(count):
						ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),proto=ip_proto,ttl=(hopcount))/"PIZZA",timeout=2)
						for s,r in ans:
							if r.haslayer(IP):
								if previousID == 0: previousID = r[IP].id
								if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
									print "IP Protocol " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								elif r.haslayer(ICMP):
									print "IP Protocol " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
									no_response_count = no_response_count + 1
								elif r.haslayer(IP):
									print "IP Protocol " + str(hopcount) + ":  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	proto=" + str(r[IP].proto)
									morehops = False
								previousID = r[IP].id
						for s in unans:
							print "IP Protocol " + str(hopcount) + ": No Response"
							no_response_count = no_response_count + 1
					if no_response_count >= max_noresp * count:
						print "IP Protocol " + str(hopcount) + ": Too many irresponsive hops. Quitting..."
						morehops = False 
						break
					print "----------------------------------------------------------------------"
					if hopcount == max_hops:
						break
					hopcount = hopcount + 1
			else:
				for i in range(count):
					ans,unans = sr(IP(dst=str(args[0]),id=RandShort(),proto=ip_proto,ttl=ip_ttl)/"PIZZA",timeout=2)
					#ans.summary(lambda (s,r): r.sprintf("IP Protocol:	ip=%IP.src% id=%IP.id% proto=%IP.proto%") )
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(IP):
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	proto=" + str(r[IP].proto)
							previousID = r[IP].id
					for s in unans:
						print "IP Protocol: No Response"
			print "----------------------------------------------------------------------"			
	except KeyboardInterrupt:
		print "Operation interrupted by the user..."
		sys.exit(1)
